//
// Created by tim on 05.04.23.
//

#include "FunctionFinderVisitor.h"
#include "ExtractionMetadata.h"
#include "Util.h"

#include <iostream>
#include <clang/AST/Mangle.h>

std::vector<std::string> getMangledName(clang::NamedDecl const *const nd) {
    if (!nd) {
        SPDLOG_ERROR("NamedDecl was nullptr");
        assert(nd && "NamedDecl and MangleContext must not be nullptr");
        return {"__NO_NAME__"};
    }

    clang::ASTNameGenerator NG(nd->getASTContext());

    if (llvm::isa<clang::CXXRecordDecl>(nd) || llvm::isa<clang::CXXMethodDecl>(nd) ||
        llvm::isa<clang::ObjCInterfaceDecl>(nd) || llvm::isa<clang::ObjCImplementationDecl>(nd)) {
        return NG.getAllManglings(nd);
    }

    return {NG.getName(nd)};
}

bool FunctionFinderVisitor::VisitFunctionTemplateDecl(clang::FunctionTemplateDecl *D) {
    if (!D->getASTContext().getSourceManager().isInMainFile(D->getSourceRange().getBegin())) {
        return true;
    }

    for (const auto &s: D->specializations()) {
        for (const auto &name: getMangledName(s)) {

            assert(callgraph->hasNode(name));

            // we should find every node in the whole program callgraph
            if (!callgraph->hasNode(name)) {
                SPDLOG_WARN("Node {} is not in the call-graph. Graph might be wrong or incomplete!", name);
                continue;
            }

            // We should only override if we have a body or if it is null
            if (s->hasBody() || callgraph->getNode(name)->getOrCreateMD<ExtractionMetadata>()->get_decl_node() == nullptr) {
                callgraph->getNode(name)->getOrCreateMD<ExtractionMetadata>()->set_decl_node(s);
            }

        }
    }

    return true;
}

bool FunctionFinderVisitor::VisitFunctionDecl(clang::FunctionDecl *functionDecl) {
    //If the function is not from main file (e.g a header defined function, don't mark
    if (!functionDecl->getASTContext().getSourceManager().isInMainFile(functionDecl->getSourceRange().getBegin())) {
        return true;
    }

    // These are the uninstantiated template descriptions (I think)
    // We traverse the instantiated once with VisitTemplateFunctionDecl
    if (functionDecl->isDependentContext()) {
        return true;
    }

    const auto &possibleNames = getMangledName(functionDecl);

    //std::cout<<functionDecl->getNameAsString()<<col2str(possibleNames," All possible manglings are:(",",",")")<<"\n";

    for (const auto &name: possibleNames) {

        assert(callgraph->hasNode(name));

        // we should find every node in the whole program callgraph
        if (!callgraph->hasNode(name)) {
            SPDLOG_WARN("Node {} is not in the call-graph. Graph might be wrong or incomplete!", name);
            continue;
        }

        // We should only override if we have a body or if it is null
        if (functionDecl->hasBody() || callgraph->getNode(name)->getOrCreateMD<ExtractionMetadata>()->get_decl_node() ==
                                       nullptr) {
            callgraph->getNode(name)->getOrCreateMD<ExtractionMetadata>()->set_decl_node(functionDecl);
        }
    }
    return true;
};

