//
// Created by tim on 21.12.23.
//
#include "PreprocessorExtract.h"


void PreprocessorExtract::If(clang::SourceLocation Loc, clang::SourceRange ConditionRange,
                             ConditionValueKind ConditionValue) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }
    assert(Loc.isValid());
    macroNestingStack.emplace(Loc, Loc, "");
}

void PreprocessorExtract::Ifdef(clang::SourceLocation Loc, const clang::Token &MacroNameTok,
                                const clang::MacroDefinition &MD) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }
    assert(Loc.isValid());
    macroNestingStack.emplace(Loc, Loc, "");
}

void PreprocessorExtract::Ifndef(clang::SourceLocation Loc, const clang::Token &MacroNameTok,
                                 const clang::MacroDefinition &MD) { Ifdef(Loc, MacroNameTok, MD); };

void PreprocessorExtract::Endif(clang::SourceLocation Loc, clang::SourceLocation IfLoc) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }
    assert(macroNestingStack.top().begin.isValid());
    assert(macroNestingStack.top().end.isValid());
    assert(IfLoc.isValid());
    assert("If these are not equal, we overwrote the last found macro end location already" &&
           macroNestingStack.top().begin == macroNestingStack.top().end);
    assert("The last found macros start should be the if loc we found" && macroNestingStack.top().begin == IfLoc);
    ExtractionBlock macro = macroNestingStack.top();
    macroNestingStack.pop();
    macro.end = Loc;
    clang::SourceRange sourceRange(macro.begin, macro.end);
    clang::CharSourceRange charSourceRange(sourceRange, true);
    macro.contents = "#" + clang::Lexer::getSourceText(charSourceRange, sourceManager, clang::LangOptions()).str();
    macros.push_back(macro);
}

inline void PreprocessorExtract::PragmaDirective(clang::SourceLocation Loc,
                                                 clang::PragmaIntroducerKind Introducer) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }
    //We only get the start of the pragma
    clang::CharSourceRange charSourceRange({Loc, Loc.getLocWithOffset(0)}, true);

    //We move forward until we encounter a newline, which finishes the pragma,
    // or enounter a \ followed by a newline, which means the pragma goes on in the next line
    int i=0;
    while(!clang::Lexer::getSourceText({{Loc, Loc.getLocWithOffset(i)}, true}, sourceManager, clang::LangOptions()).endswith("\n") ||
    clang::Lexer::getSourceText({{Loc, Loc.getLocWithOffset(i)}, true}, sourceManager, clang::LangOptions()).endswith("\\\n")  ) {
        i++;
    }
    ExtractionBlock res;
    res.begin = Loc;
    res.end = charSourceRange.getEnd();
    res.contents = clang::Lexer::getSourceText({{Loc, Loc.getLocWithOffset(i)}, true}, sourceManager, clang::LangOptions());
    pragmas.push_back(res);
}

void PreprocessorExtract::InclusionDirective(clang::SourceLocation HashLoc, const clang::Token &IncludeTok,
                                             clang::StringRef FileName, bool IsAngled,
                                             clang::CharSourceRange FilenameRange,
                                             clang::Optional<clang::FileEntryRef> File,
                                             clang::StringRef SearchPath, clang::StringRef RelativePath,
                                             const clang::Module *Imported,
                                             clang::SrcMgr::CharacteristicKind FileType) {
//Note: we currently exclude headers only by name, we could also compare them according to their sourcerange
    if (sourceManager.isInMainFile(HashLoc)) {
        includes.emplace_back(HashLoc, FilenameRange.getEnd(), "#include \"" + FileName.str() + "\"");
    }
}
